/*****************************************************************************\
 * DateOps.h
 *
 * DateOps contains misc. time and date operations
 *
 * author: mark huss (mark@mhuss.com)
 * Based in part on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

// uncomment the following line to include an assortment of non-western
// calendars:
/* #define CALENDARS_OF_THE_WORLD */


#if !defined( DATE_OPS__H )
#define DATE_OPS__H

#include "Astro.h"
#include "DateCalendars.h"
#include <time.h>
#include <stdio.h>

class DateOps {
public:

  // These are initialized in DateOps.cpp
  static const char* MONTH[];   // Month names in English
  static const char* MONTH3[];  // 3-char abbeviated month names
  static const double DST_OFFSET;

  // date format specifier.
  // examples provided are for J2000 (January 1, 2000 at noon)

  enum DATE_FMT {
      Y_MON_D,      // 2000 Jan 01
      MON_D_Y,      // Jan 01 2000
      MON_D,        // Jan 01
      Y_M_D,        // 2000-01-01
      M_D_Y,        // 01/01/2000
      M_D,          // 01/01
                    // Note: date only above, date + time below
                    //       The code depends on this separation.
      Y_MON_D_HM,   // 2000 Jan 01 12:00
      MON_D_Y_HM,   // Jan 01 2000 12:00
      MON_D_HM,     // Jan 01 12:00
      Y_M_D_HM,     // 2000-01-01 12:00
      M_D_Y_HM,     // 01/01/2000 12:00
      M_D_HM        // 01/01 12:00
  };


 /*** TIME ******************************************************************/

 /**
  * tzOffsetInDays(): calculate time zone offset from Universal Time
  *                   in days for the spec'd time_t (includes DST)
  *
  * @param tt = time_t to convert
  *
  * @return Offset (-0.5 ... +0.5 )
  */
  static double tzOffsetInDays(time_t tt);

 /**
  * tzOffsetInDays(): calculate time zone offset from Universal Time in days
  *                   at spec'd date/time (includes DST)
  *
  * @param jd - day to do calculation for
  *
  * @return Offset (-0.5 ... +0.5 )
  */
  static double tzOffsetInDays(double jd) {
    return tzOffsetInDays( dayToTime(jd) );
  }

 /**
  * tzOffsetInDays(): calculate time zone offset from Universal Time
  *                   in days using the current time (includes DST)
  *
  * @return Offset (-0.5 ... +0.5 )
  */
  static double tzOffsetInDays() {
    return tzOffsetInDays( time(0) );
  }

 /**
  * dstOffsetInDays(): calculate daylight savings time offset in days at
  *                    the spec'd time
  *
  * @param time_t - date to use
  *
  * @return Offset ( 0 or 1/24 )
  */
 static double dstOffsetInDays(time_t tt);

 /**
  * dstOffsetInDays(): calculate daylight savings time offset in days at
  *                    the spec'd time
  *
  * @param jd - date to use
  *
  * @return Offset ( 0 or 1/24 )
  */
 static double dstOffsetInDays(double jd) {
   return dstOffsetInDays( dayToTime(jd) );
 }

 /**
  * dstOffsetInDays(): calculate current daylight savings time offset in days
  *
  * @return Offset ( 0 or 1/24 )
  */
  static double dstOffsetInDays() {
   return dstOffsetInDays( time(0) );
 }

 /**
  * formatTime(): format a time into an HH:MM string
  *
  * @param buf - where to put the string
  * @param dayFrac - a fractional day ( >= 0.0, < 1.0 )
  * @param doSecs - true to include seconds (HH:MM:SS)
  *
  * @return 1 if rounding wrapped to 00:00[:00], else 0
  */
  static int formatTime( char* buf, double dayFrac, bool doSecs = false );

 /**
  * formatTime(): format a time into an HH:MM string, and write it to a FILE
  *
  * @param fp - where to write the string
  * @param dayFrac - a fractional day ( >= 0.0, < 1.0 )
  * @param doSecs - true to include seconds (HH:MM:SS)
  *
  * @return 1 if rounding wrapped to 00:00[:00], else 0
  */
  static int formatTime( FILE* fp, double dayFrac, bool doSecs = false );

 /**
  * dayToHMS(): break the fractional part of a Julian day into hours, minutes,
  * and seconds
  *
  * @param h - where to put the hour
  * @param m - where to put the minute
  * @param s - where to put the second
  * @param dayFrac - a fractional day ( >= 0.0, < 1.0 )
  */
  static void dayToHMS(int& h, int& m, int& s, double jd);

 /**
  * dayToHM(): break the fractional part of a Julian day into hours and minutes
  *
  * @param h - where to put the hour
  * @param m - where to put the minute
  * @param dayFrac - a Julian Day
  *
  * @return 1 if rounding wrapped to 00:00, else 0
  */
  static int dayToHM(int& h, int& m, double jd);

 /**
  * hourToDay(): convert hour to decimal day, e.g., 12h -> .5d
  *
  * @param hour
  *
  * @return day fraction
  */
  static double hourToDay( int hours ) { return double(hours) / Astro::HOURS_PER_DAY; }

 /**
  * formatDateTime(): format a JD into a text string
  *
  * @param buf - where to put the formatted string
  * @param jd  - the day to format
  * @param fmt - format type (see DateOps::DATE_FMT)
  */
  static void formatDateTime( char* buf, double jd, DATE_FMT fmt );

 /**
  * formatDateTime(): format a JD into a text string and write it to a FILE
  *
  * @param fp - where to write the formatted string
  * @param jd  - the day to format
  * @param fmt - format type (see DateOps::DATE_FMT)
  */
  static void formatDateTime( FILE* fp, double jd, DATE_FMT fmt );

 /**
  * formatMS(): format a fractional minute into a text string (MM:SS.S)
  *
  * @param buf - where to put the formatted string
  * @param m  - the value (in minutes) to format
  */
  static void formatMS( char* buf, double min ) {
    sprintf( buf, "%02d:%02.1f\n",
             int(min), (min - int(min)) * Astro::SECONDS_PER_MINUTE );
  }

 /**
  * formatMS(): format a fractional minute into a text string (MM:SS.S) and
  *             write it the spec'd file.
  *
  * @param fp - where to write the formatted string
  * @param m  - the value (in minutes) to format
  */
  static void formatMS( FILE* fp, double min ) {
    fprintf( fp, "%02d:%02.1f\n",
             int(min), (min - int(min)) * Astro::SECONDS_PER_MINUTE );
  }

 /**
  * timeToLDay(): convert a "local" time_t to Julian Day
  *
  * @param time - time_t to convert
  *
  * @return Julian Day
  */
  static double timeToLDay( time_t time ) {
    return timeToDay( localtime(&time) );
  }

 /**
  * timeToUDay(): convert UTC time_t to Julian Day
  *
  * @param time_t to convert
  *
  * @return Julian Day
  */
  static double timeToUDay( time_t time ) {
    return timeToDay( gmtime(&time) );
  }

 /**
  * dayToTime(): convert a JD to "local" time_t (including DST)
  *
  * @param jd - jd to convert
  *
  * @return equivalent time_t
  */
  static time_t dayToTime( double jd );

 /**
  * now(): Determine the Julian Day value at the present moment
  *
  * @param local - true for local time, false for UTC
  *
  * @return The Julian Day value
  */
  static double now(bool local = false) {
    return local ? timeToLDay( time(0) ) : timeToUDay( time(0) );
  }

 /**
  * greenwichSiderealTime(): convert a "local" time_t to GST
  *
  * @param jd - julian day
  *
  * @return apparent Greenwich sidereal time (in Radians) for the given jd
  */
  static double greenwichSiderealTime( double jd );

 /*** DATE ******************************************************************/

 /**
  * toJulianCenturies(): convert a JD to Julian Century referenced to epoch
  *     J2000. This is the number of days since J2000 converted to centuries.
  *
  * @param Julian Day Number
  *
  * @return centuries since J2000 (12 noon on January 1, 2000)
  */
  static double toJulianCenturies( double jd ) {
    return ( jd - Astro::J2000 ) / Astro::DAYS_PER_CENTURY;
  }

 /**
  * toJulianMillenia(): convert a JD to Julian Millenia referenced to epoch
  *     J2000. This is the number of days since J2000 converted to millenia.
  *
  * @param Julian Day Number
  *
  * @return millenia since J2000 (12 noon on January 1, 2000)
  */
  static double toJulianMillenia( double jd ) {
    return ( jd - Astro::J2000 ) / Astro::DAYS_PER_MILLENIUM;
  }

 /**
  * dmyToDay(): convert a day/month/year to a long Julian Day
  *
  * @param day - day of the month
  * @param month - month of the year
  * @param year - year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  *
  * @return equivalent Julian Day rounded to a long
  */
  static long dmyToDay(
      int day, int month, int year, CalendarType calendar = T_GREGORIAN );

 /**
  * dmyToDoubleDay(): convert a day/month/year to a double Julian Day
  *
  * @param day - day of the month
  * @param month - month of the year
  * @param year - year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  *
  * @return equivalent Julian Day as a double
  */
  static double dmyToDoubleDay(
      int day, int month, int year, CalendarType calendar = T_GREGORIAN )
  {
    return /*Astro::JD_GREG_DIFF +*/ double(dmyToDay( day, month, year, calendar ));
  }

 /**
  * dmyToDoubleDay(): convert a day/month/year to a double Julian Day
  *
  * @param day - day of the month
  * @param month - month of the year
  * @param year - year
  * @param second - second in the day ( 0 .. Astro::SECONDS_PER_DAY)
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  *
  * @return equivalent Julian Day as a double
  */
  static double dmyToDoubleDay(
      int day, int month, int year, int second, CalendarType calendar = T_GREGORIAN )
  {
    return /*Astro::JD_GREG_DIFF +*/ double(second)/Astro::SECONDS_PER_DAY +
        dmyToDay( day, month, year, calendar );
  }

 /**
  * dayToDmy(): convert a long Julian Day to day/month/year
  *
  * @param jd - Julian Day to convert
  * @param day& - where to put the day of the month
  * @param month& - where to put the month of the year
  * @param year& - where to put the year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  */
  static void dayToDmy(
      long jd, int& day, int& month, int& year, CalendarType calendar = T_GREGORIAN );

  static void dayToDmy(
      double jd, int& day, int& month, int& year, CalendarType calendar = T_GREGORIAN )
  {
    return dayToDmy( long(floor(jd)), day, month, year, calendar);
  }
 /**
  * dayToDmy(): convert a long Julian Day to day/month/year
  *
  * @param jd - Julian Day to convert
  * @param day* - where to put the day of the month
  * @param month* - where to put the month of the year
  * @param year* - where to put the year
  * @param calendar - (optional) T_GREGORIAN or T_JULIAN, former is the default
  */
  static void dayToDmy(
      long jd, int* day, int* month, int* year, CalendarType calendar = T_GREGORIAN )
  {
    dayToDmy( jd, *day, *month, *year, calendar );
  }

  static void dayToDmy(
      double jd, int* day, int* month, int* year, CalendarType calendar = T_GREGORIAN )
  {
    dayToDmy( long(floor(jd)), *day, *month, *year, calendar );
  }
 //--------------------------------------------------------------------------
 /**
  * dstStart(): Determine the Julian Day in the specified year where Daylight
  *             Savings Time starts
  *
  * @param The year
  *
  * @return The Julian Day value
  */
  static long dstStart(int year);

 /**
  * dstEnd(): Determine the Julian Day in the specified year where Daylight
  *           Savings Time ends
  *
  * @param The year
  *
  * @return The Julian Day value
  */
  static long dstEnd(int year);

private:

 /**
  * roundToNearestMinute(): round a time to the nearest minute
  *
  * if h:m wraps to zero, we roll back to 23:59 to preserve "today"
  *
  * @param h - hour
  * @param m - minute
  * @param s - second
  *
  * @return 1 if rounding wrapped to 00:00, else 0.
  *
  */
  static int roundToNearestMinute(int &h, int& m, int&s);

  static double timeToDay( struct tm* pt );

};

#endif  /* #if !defined( DATE_OPS__H ) */

