/*****************************************************************************\
 * Lunar.h
 *
 * Lunar is a class that can calculate lunar fundmentals for any reasonable
 * time.
 *
 * author: mark huss (mark@mhuss.com)
 * Based on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

#if !defined( LUNAR__H )
#define LUNAR__H

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include "AstroOps.h"
#include "DateOps.h"
#include "MathOps.h"

class ObsInfo;

//-------------------------------------------------------------------------
// A struct to hold the fundmental elements
// The member names should be familiar to Meeus fans ;-)
//
struct LunarFundamentals {
  double Lp;  // moon's mean longitude
  double D;   // moon's mean elongation
  double M;   // sun's mean anomaly
  double Mp;  // moon's mean anomaly
  double F;   // mean distance of the Moon from its ascending node
  double A1;  // further argument 1
  double A2;  // further argument 2
  double A3;  // further argument 3
  double T;   // time in centuries since J2000
  double T2;  // T * T

  LunarFundamentals():Lp(0.), D(0.), M(0.), Mp(0.), F(0.),
                      A1(0.), A2(0.), A3(0.), T(0.), T2(0.)
  {}
};

//-------------------------------------------------------------------------
// our main class -- calculates Lunar fundamentals, lat, lon & distance
//
class Lunar {
public:
  enum Quarter { NEW, Q1, FULL, Q3 };

  // default c'tor
  Lunar() :
      m_initialized( false ),
      m_jd( Astro::INVALID ),
      m_lon( Astro::INVALID ),
      m_lat( Astro::INVALID ),
      m_r( Astro::INVALID )
  {}

  // data & time c'tor
  // js = Julian Day
  Lunar( double jd ) {
    calcFundamentals( jd );
  }

  // Exact number of days from one new moon to the next new moon.
  static const double SYNODIC_MONTH = 29.530588861;
  static const double SYNODIC_WEEK = /*SYNODIC_MONTH*/ 29.530588861/4.0;

 /**
  * illuminatedFraction() - calculate the illuminated fraction of the lunar
  *                         surface using the data in LunarFundamentals (m_f)
  *
  * @return fraction illuminated (0.0...0.999), or -1 if class has not been
  *         initialized.
  */
  double illuminatedFraction() {
    return m_initialized ?
           (1.0 + cos(phaseAngle()) ) / 2. :
           Astro::INVALID;
  }

 /**
  * ageOfMoonInDays() - Calculate the age of the moon in days
  *                     (0.0 ... SYNODIC_MONTH)
  *
  * @param jd - Julian Day to calculate for
  *
  * @return - the age in days
  */
  static double ageOfMoonInDays( double jd );

 /**
  * ageOfMoonInDays() - Calculate the age of the moon in days
  *                     (0.0 ... SYNODIC_MONTH)
  *
  * @param luna - existing Lunar instance to use with embedded JD
  *
  * @return - the age in days
  */
  static double ageOfMoonInDays( Lunar& luna );

 /**
  * ageOfMoonInDays() - Calculate the age of the moon in days
  *                     (0.0 ... SYNODIC_MONTH) for the current time
  *
  * @return - the age in days
  */
  static double ageOfMoonInDays( /* now */ ) {
    return ageOfMoonInDays( DateOps::now() );
  }

 /**
  * ageOfMoonInDays() - Calculate the age of the moon in days
  *                     (0.0 ... SYNODIC_MONTH)
  *
  * @param day - day to use
  * @param month - month to use
  * @param year - year to use
  * @param second - second of the day to use
  *
  * @return - the age in days
  */
  static double ageOfMoonInDays( int day, int month, long year, long second ) {
    return ageOfMoonInDays( DateOps::dmyToDoubleDay( day, month, year, second ) );
  }

 /**
  * getPhase(): Calculate the time of the quarter nearest the given Julian
  * day, using Meeus' algorithm. Input JD must be within SYNODIC_MONTH/2 days
  * of actual event, else this will return the corresponding event from
  * another lunation.
  *
  * @param jd A Julian day close to the expected date
  * @param quarter (<TT>Lunar.NEW</TT>, <TT>Q1</TT>, <TT>FULL</TT>,
  *     or <TT>Q3</TT>)
  * @param localTZ - true to convert return value to local TZ
  *
  * @return The relatively exact Julian day with decimal
  */
  static double getPhase( long jd, Quarter quarter, bool localTZ=true );

 /**
  * getPhase(): Calculate the time of the quarter nearest the given Julian
  * day, using Meeus' algorithm. Input JD must be within SYNODIC_MONTH/2 days
  * of actual event, else this will return the corresponding event from
  * another lunation.
  *
  * @param jd A Julian day close to the expected date
  * @param quarter (<TT>Lunar.NEW</TT>, <TT>Q1</TT>, <TT>FULL</TT>,
  *     or <TT>Q3</TT>)
  * @param localTZ - true to convert return value to local TZ
  *
  * @return The relatively exact Julian day with decimal
  */
  static double getPhase( double jd, Quarter quarter, bool localTZ=true ) {
    return getPhase( long(floor(jd)), quarter, localTZ );
  }

 /**
  * lunation(): return the number in Ernest W. Brown's numbered series
  *             of lunar cycles for the specified JD.
  *
  * The base Julian day for Ernest W. Brown's numbered series of
  * lunations = 1923 Jan 17 02:41 UT.
  * <P>
  * This date has been widely quoted as "Jan 16 1923" and indeed it
  * was (in EST) at Yale University where Prof. Brown worked.
  *
  * @param jd Julian day
  *
  * @return Lunation number
  */
  static const double LUNATION_BASE = 2423436.40347;

  static int lunation( double jd );

 /**
  * lunation(): return the number in Ernest W. Brown's numbered series
  *             of lunar cycles for the specified JD.
  *
  * The base Julian day for Ernest W. Brown's numbered series of
  * lunations = 1923 Jan 17 02:41 UT.
  * <P>
  * This date has been widely quoted as "Jan 16 1923" and indeed it
  * was (in EST) at Yale University where Prof. Brown worked.
  *
  * @param jd Julian day
  *
  * @return Lunation number
  */
  static int lunation( long jd ) {
    return lunation( double( jd ) );
  }

 /**
  * lunation(): return the number in Ernest W. Brown's numbered series
  *             of lunar cycles for the current time.
  *
  * The base Julian day for Ernest W. Brown's numbered series of
  * lunations = 1923 Jan 17 02:41 UT.
  * <P>
  * This date has been widely quoted as "Jan 16 1923" and indeed it
  * was (in EST) at Yale University where Prof. Brown worked.
  *
  * @return Lunation number
  */
  static int lunation( /* now */ ) {
    return lunation( DateOps::now() );
  }

 /**
  * summary(): Format a text summary of current lunar information, including
  *            the lunation, age, rise/set times and time of full/new moons
  *            when appropriate.
  *
  * @param age - where to put the age in days of the moon.
  * @param lun - where to put the lunation
  * @param buf - where to put the formatted string
  * @param oi - location information for the observer
  * @param doHtml - if true, format text in HTML form, else plain ASCII
  *
  */
  static void summary(int& age, int& lun, char* buf,
                      ObsInfo& oi, bool doHtml = false);

 /**
  * nextMoon(): return the approximate JD of the next new moon after the
  *             specified JD. If no JD is specified, uses now() as the
  *             starting point.
  *
  * @param jd - Julian Day to start
  *
  * @return Julian Day of next new moon
  */
  static double nextMoon(double jd = -1.0);

 /**
  * calcAllLocs() - calculate all three location elements of the moon at the
  *                 specified time
  *
  * @param lon - where to put the longitude
  * @param lat - where to put the latitude
  * @param rad - where to put the radius
  * @param jd  - time to calculate location elements for
  */
  static void calcAllLocs(
      double& lon,            // returned longitude
      double& lat,            // returned latitude
      double& rad,            // returned radius vector
      double jd);             // time

 /**
  * calcFundimentals() - calculates the lunar fundamanentals for the given time
  *                      must be called before using most non-static members
  *                      of this class.
  *
  * @param jd - Julian Day
  */
  void calcFundamentals( double jd );

  //
  // NOTE: calcFundamentals() must be called before calling the functions
  //       below, or an invalid result (-1.) will be returned.
  //

 /**
  * phaseAngle() - returns the phase angle of the moon at the current time
  *
  * @return phase angle
  */
  double phaseAngle();

 /**
  * jd() - getter for instance lunar JD
  *
  * @return - Julian Day used to calculate the fundimental data in this
  *           instance
  */
  double jd() { return m_jd; }  // returns -1 if not initialized

 /**
  * latitude() - getter for instance lunar latitude
  *
  * @return - latitude in degrees
  */
  double latitude();         // returns -1 if not initialized

 /**
  * latitudeRadians() - getter for instance lunar latitude
  *
  * @return - latitude in radians
  */
  double latitudeRadians() {  // returns -1 if not initialized
    return ( m_initialized ) ? MathOps::toRadians( latitude() ) : -1.;
  }

 /**
  * longitude() - getter for instance lunar longitude
  *
  * @return - latitude in degrees
  */
  double longitude()         // returns -1 if not initialized
  {
    if ( m_lon < 0. )
      calcLonRad();
    return m_lon;
  }

 /**
  * longitudeRadians() - getter for instance lunar longitude
  *
  * @return - latitude in radians
  */
  double longitudeRadians()  // returns -1 if not initialized
  {
    return ( m_initialized ) ? MathOps::toRadians( longitude() ) : -1.;
  }

 /**
  * radius() - getter for instance lunar radius (distance)
  *
  * @return - distance
  */
  double radius()     // returns -1 if not initialized
  {
    if ( m_r < 0. )
      calcLonRad();
    return m_r;
  }

 //---------------------------------------------------------------------------

private:

  // calculate an individual lon/lat/rad fundimental
  //  tptr - points to array of doubles
  //  t - time in decimal Julian centuries
  //
  double getFund( const double* tptr, double t );

  // calculate an individual phase fundimental
  //  tptr - points to array of doubles
  //  k - phase constant, 0 = new moon Jan 6, 2000
  //  t - time in decimal Julian centuries
  //
  static double getFund( const double* tptr, double k, double t );

  // calculate longitude and radius
  //
  // NOTE: calcFundamentals() must have been called first
  //
  void calcLonRad();

  // ***** data  *****

  // our calculated fundmentals
  //
  LunarFundamentals m_f;

  // true if calcFundamentals has been called
  bool m_initialized;

  // JD used to initialize instance
  double m_jd;

  // longitude, latitude, and radius (stored in _degrees_)
  double m_lon, m_lat, m_r;
};

#endif  /* #if !defined( LUNAR__H ) */

