/*****************************************************************************\
 * Vsop.cpp
 *
 * The Vsop class wraps the VSOP87 data and provides VSOP support fns
 *
 * author: mark huss (mark@mhuss.com)
 * Based on Bill Gray's open-source code at projectpluto.com
 *
\*****************************************************************************/

#include "Vsop.h"
#include "VsopData.cpp"
#include "Astro.h"
#include "MathOps.h"
#include "DateOps.h"
#include <math.h>

/*
 * This function, using the simplified VSOP87 data in Meeus, can compute
 * planetary positions in heliocentric ecliptic coordinates.
 *
 * 'planet' can run from 0=sun,  1=mercury,  ... 8=neptune.
 * (VSOP doesn't handle the moon or Pluto.)
 *
 * 'value' can be either
 *   0=ecliptic longitude, 1=ecliptic latitude, 2=distance from sun.
 *   (These are ecliptic coordinates _of date_,  by the way!)
 *
 * cen = (JD - 2451545.) / 36525. = difference from J2000,  in Julian
 * centuries.
 *
 * returns 0. if planet is <1 (mercury) or >8 (neptune)
 */
double Vsop::calcLoc(
    double cen,         // time in decimal centuries
    Planet planet,
    LocType ltype)
{
  double rval = 0.0;

  if( planet >= SUN && planet < PLUTO && ltype >= ECLIPTIC_LON && ltype <= RADIUS ) {

    if ( SUN == planet )
      planet = EARTH;

    double jm = cen / 10.;          // convert to julian millenia
    double tPower = 1.0;

    const VsopTerms* pT;
    switch ( planet ) {
      case MERCURY: pT = &(MercuryTerms[ltype * Vsop::TERM_SETS]); break;
      case VENUS:   pT = &(  VenusTerms[ltype * Vsop::TERM_SETS]); break;
      case EARTH:   pT = &(  EarthTerms[ltype * Vsop::TERM_SETS]); break;
      case MARS:    pT = &(   MarsTerms[ltype * Vsop::TERM_SETS]); break;
      case JUPITER: pT = &(JupiterTerms[ltype * Vsop::TERM_SETS]); break;
      case SATURN:  pT = &( SaturnTerms[ltype * Vsop::TERM_SETS]); break;
      case URANUS:  pT = &( UranusTerms[ltype * Vsop::TERM_SETS]); break;
      case NEPTUNE: pT = &(NeptuneTerms[ltype * Vsop::TERM_SETS]); break;
      default: break;
    };

    // Always six series to calculate
    for( int i=0; i<6; i++ ) {
      double sum = 0.;
      const VsopSet* pv = pT->pTerms;

      // sum the term = A x cos( B + C x tc ) for each row
      for( unsigned j=0; j<pT->rows; j++ ) {
        sum += pv->A * cos( pv->B + pv->C * jm );
        pv++;
      }
      // Add to series and bump multipler
      // i.e., L = L0*t + L1*t^2 + L2*t^3 + ...
      rval += sum * tPower;
      tPower *= jm;
      pT++;
    }

    rval *= 1.e-8;  // rescale the term

    if( ECLIPTIC_LON == ltype )  /* ensure 0 < rval < 2PI  */
      rval = MathOps::normalizeRadians( rval );
  }

  return rval;
}

void Vsop::calcAllLocs(
    double& lon,            // returned longitude
    double& lat,            // returned latitude
    double& rad,            // returned radius vector
    double jd,              // Julian Day
    Planet planet)          // must be in the range SUN...NEPTUNE
{
  // t is in julian centuries from J2000.0
  double t = DateOps::toJulianCenturies( jd );

  lon = calcLoc( t, planet, ECLIPTIC_LON );
  lat = calcLoc( t, planet, ECLIPTIC_LAT );
  rad = calcLoc( t, planet, RADIUS );
}

